#  This file is part of BGX, the Bayesian Gene eXpression program.
#  Copyright 2008 Ernest Turro <ernest.turro@ic.ac.uk>
#
#  BGX is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License, version 2, as
#  published by the Free Software Foundation.
#
#  BGX is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


"mmbgx" <-
function(aData, arrayType=NULL, ensemblVersion=NULL, geneLevel=FALSE, burnin=8192, iter=16384, inputDirs=tempfile(paste("input", c(1:length(aData)), sep="_")), outputDir=".", standalone=FALSE) {

  if(! extends(class(aData), "FeatureSet") )
    stop("Please read in your CEL files with read.celfiles() from the oligo package.")

  if(is.null(arrayType)) {
    stop("Please specify the array type. Try:\nlibrary(affyio)\nread.celfile.header(path_to_cel_file)$cdfName.\n")
  }

  if(ncol(aData) != length(inputDirs)) {
    stop("ncol(aData) != length(inputDirs).\n")
  }

  firstData <- aData[,1] # one array at a time
  inputDir <- inputDirs[1]

  if(geneLevel==TRUE && length(grep("Gene", arrayType))==0) {
    cat("Analyse array", sampleNames(firstData), "at the Ensembl gene level.\n")
    if(length(grep("-gene", arrayType))==0)
      arrayType=paste(arrayType,"-gene",sep="")
  } else {
    cat("Analyse array", sampleNames(firstData), "at the Ensembl transcript level.\n")
  }

  if(burnin %% 1024 != 0 || iter %% 1024 != 0 || burnin<=0 || iter<=0)
    stop("\"iter\" and \"burnin\" must be positive multiples of 1024")


  availVersions = dir(path.expand(file.path(.path.package("mmbgx"), "structDirs",arrayType)))

  if(length(availVersions)==0)
    stop("No structure files available for ", arrayType, ". Try creating them with makeStructDir()")
  if(length(availVersions) > 1) {
    if(!is.null(ensemblVersion) && ! ensemblVersion %in% availVersions) {
      stop("The Ensembl version specified is not available. The following are available: ", paste(availVersions, collapse=", "))
    } 
  }
  if(is.null(ensemblVersion))
    ensemblVersion=availVersions

  structDir = path.expand(file.path(.path.package("mmbgx"), "structDirs",arrayType,ensemblVersion))
  if(!file.exists(structDir)) 
    stop("Sorry, no structure files are available for the array type ", arrayType,
      ". The following array types are supported: ",
      paste(dir(file.path(.path.package("mmbgx"), "structDirs")),collapse=", "), ".\n") 

  inputDir = path.expand(inputDir)
  outputDir = path.expand(outputDir)

  # create input dir where input files will be saved
  if(.Platform$OS.type=="windows") inputDir <- gsub("\\\\","/",inputDir) # Use / as file separator
  if(file.exists(inputDir)) {
    cat("Warning: input directory '", inputDir, "' already exists.\nOverwrite? (y/N) ", sep="")
    ans = scan(what=character(0), nmax=1, quiet=TRUE)
    if(length(ans)==0 || tolower(ans) != 'y') stop("Aborting.")
  }
  dir.create(inputDir, showWarnings = FALSE, recursive =TRUE)

  # create directory where runs will be saved if necessary
  if(.Platform$OS.type=="windows") outputDir <- gsub("\\\\","/",outputDir) # Use / as file separator
  if(!file.exists(outputDir)) dir.create(outputDir)

  # Copy structure files to input directory (PM2PS.txt, PS2PM.txt, etc...)
  file.copy(file.path(structDir,dir(structDir)), inputDir, overwrite=TRUE)

  # get number of probesets 
  nPS <- length(scan(file.path(inputDir,"PSids.txt"), quiet=TRUE, what=character(0)))

  numArrays<-ncol(firstData)
  noOfProbes <- 0
  # write PM file
  message("Writing input files to ", as.character(inputDir))
  res <- .C("writePMs",as.integer(oligo:::pm(firstData)), as.integer(nrow(oligo:::pm(firstData))), as.integer(numArrays), as.character(file.path(inputDir, "PMids.txt")), as.character(file.path(inputDir, "PM.txt")), as.integer(noOfProbes))
  noOfProbes <- res[[6]]
#  PMids <- scan(file.path(inputDir,"PMids.txt"), quiet=TRUE)
#  write(exprs(aData)[PMids,], file=file.path(inputDir,"PM.txt"))

  # write gamma and delta2 files
  BGids <- scan(file.path(inputDir,"BGids.txt"), quiet=TRUE)
  BGcats <- scan(file.path(inputDir, "BGcats.txt"), quiet=TRUE)

  gamma <- delta2 <- c()

  for(c in 1:numArrays) {
    for(i in 0:25){
      if(any(BGcats==i)){ # Check if there are any BG probes with this GC content
        gamma[(c-1)*26 + i+1] <- mean(log(oligo:::pm(firstData)[BGids[BGcats==i],c]))
        delta2[(c-1)*26 + i+1] <- var(log(oligo:::pm(firstData)[BGids[BGcats==i],c]))
      } else {
        gamma[(c-1)*26 + i+1] <- delta2[(c-1)*26 + i+1] <- 0
      }
    }
  }

  write(gamma, file=file.path(inputDir,"gamma.txt"))
  write(delta2, file=file.path(inputDir,"delta2.txt"))

  # write sample sets file
  # use of sampleSets obsolete; we only analyse one array at a time
  # so set sampleSets=1
  sampleSets=1
  if(is.null(sampleSets)){
    if(!is.null(pData(firstData)[,1])) {
      sampleSets <- c()
      cat("Note: sampleSets not specified. Using pData(AffyBatch)[,1] to group samples into conditions.\n")
      for(con in levels(factor(pData(firstData)[,1]))) {
        sampleSets <- append(sampleSets, sum(pData(firstData)[,1] == con))
      }
    } else {
      sampleSets=rep(1,numArrays)
    }
  } else if(sum(sampleSets)!=numArrays)
    stop("sampleSets not sensible.  You have ", numArrays, " arrays, but sum(sampleSets) = ",sum(sampleSets))
#  cat("Analysing",numArrays,"array(s) across", length(sampleSets), "condition(s):\n")
#  for(a in 1:length(sampleSets)){
#    cat("\t- condition ",a,": ",sampleSets[a], " array(s)\n",sep="")
#  }
  write(sampleSets,file.path(inputDir,"SS.txt"))

  # files to copy to output directory
  fil <- c("PSids.txt", "geneIDs.txt", "geneTranscripts.txt")
  file.create(file.path(inputDir,"filesToCopy.txt"))
  for(f in fil) {
    if(file.exists(file.path(inputDir,f))) write(f, file.path(inputDir,"filesToCopy.txt"), append=TRUE)
  }

  # write infile.txt
   cat(paste("samples ",numArrays,"\n",
    "conditions ",length(sampleSets),"\n",
    "probes ", noOfProbes,"\n",
    "probesets ",nPS,"\n",
    "PM2PS PM2PS.txt\n",
    "PS2PM PS2PM.txt\n",
    "filesToCopy filesToCopy.txt\n",
    "SampleSets SS.txt\n",
    "Categories PMcats.txt\n",
    "numberOfCategories ", length(gamma)/numArrays, "\n",
    "PM PM.txt\n",
    "gamma gamma.txt\n",
    "delta2 delta2.txt\n",
    "seed ",192492,"\n",
    "sweeps ",iter,"\n",
    "burn-in ",burnin,"\n",
    "Adaptive 1\n",
    "BatchSize 50\n",
    "OptimalAR 0.44\n",
    "S_jmp 30\n",
    "Mu_jmp 1.1\n",
    "Sigma_jmp 1.5\n",
    "CEL_file ", sampleNames(firstData), "\n", sep=""),
    file=file.path(inputDir,"infile.txt"))

    cat("MMBGX input files are in \"",inputDir,"\".\n",sep="")

    if(!standalone) {
      # free allocated memory on user interrupt/end of simulation
      on.exit(.C("freeBGXMemory", PACKAGE = "mmbgx"))
   
      res <- .C("mmbgxFromR",as.character(file.path(inputDir,"infile.txt")), as.character(outputDir));
    }
    
    if(length(aData)>1) {
      mmbgx(aData[,2:length(aData)], arrayType,geneLevel, burnin, iter, inputDirs[2:length(aData)], outputDir, standalone)
    }
}

  

##########################################
##########################################

# analysis functions

##########################################
##########################################




### Read in MMBGX output

readSingle.mmbgx <- function(path) { 
  res <- list()
  cat("Reading '",path, "'\n", sep="")
  summary <- read.delim(file.path(path,"summary.txt"), row.names=1)

  for(i in 1:nrow(summary)) { cat("***", row.names(summary)[i], "\n   ", paste(summary[i,1]),"\n")}
  
#  summary <- as.vector(summary[[1]])
  noOfConditions <- as.numeric(as.vector(summary["Number of conditions:",]))
  noOfProbesets <- as.numeric(as.vector(summary["Number of probesets:",]))
  subsample <- as.numeric(as.vector(summary["Subsampling interval:",]))
  iter <- as.numeric(as.vector(summary["Number of sampling sweeps:",]))
  traceLength <- iter/subsample
  
  res$mu <- list() 
  for(m in 1:noOfConditions) {
    cat("Reading mu under condition ", m, "...", sep="")
    res$mu[[m]] <- matrix(scan(file.path(path, paste("mu.",m,sep="")), what=double(0)), ncol=traceLength, nrow=noOfProbesets)
  }
  cat("Reading mu average...")
  res$muave <- matrix(scan(file.path(path, "muave"), what=double(0)), ncol=noOfConditions, nrow=noOfProbesets)

    cat("Reading transcript IDs...")
  res$PSids <- scan(file.path(path, "PSids.txt"), what="character")
  if(file.exists(file.path(path, "geneIDs.txt"))) {
    cat("Reading gene IDs...")
    res$geneIDs <- scan(file.path(path, "geneIDs.txt"), what="character")
  }
  if(file.exists(file.path(path, "geneTranscripts.txt"))) {
    cat("Reading gene-transcript mappings...")
    res$geneTranscripts <- readVarLines(file.path(path, "geneTranscripts.txt"))
  }
  cat("Done.\n")
  return(res)
}

combineRuns.mmbgx <- function(..., sampleSets=NULL, outputDir=NULL) {

  # validate arguments and compatibility of run dirs 
  if(is.null(outputDir)) stop("Please specify an outputDir.\n")
  args <- list(...)
  paths <- c()
  for(i in 1:length(args)){
    for(j in 1:length(args[[i]])){
      paths <- append(paths,args[[i]][j])
    }
  }
  if(length(paths) <= 1)
    stop("No directories to combine.\n")
  if(sum(sampleSets)!=length(paths))
    stop("sampleSets not sensible.  You have ", length(paths), " output paths, but sum(sampleSets) = ",sum(sampleSets))
  PSids <- NULL

  tracelens <- c()
  for(path in paths){
    pis <- scan(file.path(path, "PSids.txt"), what="character", quiet=TRUE)
    PSids <- cbind(PSids,pis)
    cursum <- read.delim(file.path(path,"summary.txt"), row.names=1)
    subsample <- as.numeric(as.vector(cursum["Subsampling interval:",]))
    iter <- as.numeric(as.vector(cursum["Number of sampling sweeps:",]))
    tracelens <- append(tracelens, iter/subsample)
  }
  for(i in nrow(PSids)){
    if(any(PSids[i,1] != PSids[i,])) stop("Sorry, incompatible run directories: probeset IDs do not match.\n")
  }
  if(any(tracelens[1] != tracelens)) stop("Sorry, incompatible run directories: trace lengths do not match.\n")

  if(.Platform$OS.type=="windows") outputDir <- gsub("\\\\","/", outputDir) # Use / as file separator
  if(file.exists(outputDir)) {
    cat("Warning: output directory '", outputDir, "' already exists.\nOverwrite? (y/N) ", sep="")
     ans = scan(what=character(0), nmax=1, quiet=TRUE)
     if(length(ans)==0 || tolower(ans) != 'y') stop("Aborting.")
  }

  # create output dir and copy files 
  dir.create(outputDir, showWarnings = FALSE, recursive =TRUE)
  file.copy(file.path(path,"PSids.txt"), file.path(outputDir))
  if(file.exists(file.path(paths[1], "geneIDs.txt")) && file.exists(file.path(paths[1], "geneTranscripts.txt"))) {
    file.copy(file.path(paths[1],"geneIDs.txt"), file.path(outputDir))
    file.copy(file.path(paths[1],"geneTranscripts.txt"), file.path(outputDir))
  }

  # loess normalise, average within conditions and write 
  muave <- c()
  for(p in paths) {
    muave <- cbind(muave,scan(file.path(p,"muave"), quiet=TRUE))
  }
  mmuave <- apply(muave,1,mean)
  loessmat <- matrix(nrow=length(mmuave), ncol=length(paths))
 
  sstrans <- floor(seq(1, length(mmuave), length.out=20000)) # for speed, apply to 20000 points, predict others
  lowest100 <- order(mmuave)[1:100]
  highest100 <- order(mmuave,decreasing=T)[1:100]
  sstrans <- sort(union(sstrans, lowest100)) # add the extremes too, as these can result in NAs in predict()
  sstrans <- sort(union(sstrans, highest100))
  for(r in 1:length(paths)) {
    cat("Loess normalising array ", r, "/", length(paths),"\r")

    muaveSubset <- data.frame(M = muave[sstrans,r] - mmuave[sstrans], A = mmuave[sstrans])
    muaveSubsetLoess <- loess(M~A, muaveSubset)
    loessmat[,r] <- predict(muaveSubsetLoess, data.frame(A=mmuave),se=FALSE)
    if(any(is.na(loessmat[,r]))) stop("NAs predicted in loess.predict.")
    muave[,r] <- muave[,r] - loessmat[,r]
  }




  cat("\nAveraging within conditions")
  
  condmuave <- c()
  i <- 1
  for(s in sampleSets) {
    if(s>1) {
      condmuave <- cbind(condmuave, apply(muave[,i:(i+s-1)], 1,mean))
    } else {
      condmuave <- cbind(condmuave, muave[,i])
    }
    i <- i+s
  }
  write(condmuave, file=file.path(outputDir, "muave"), ncolumns=nrow(condmuave))

  .Call("loessave",
    as.character(file.path(paths,"mu.1")),
    as.integer(ncol(muave)),
    as.integer(sampleSets),
    as.integer(length(sampleSets)),
    as.double(loessmat),
    as.integer(nrow(muave)),
    as.integer(tracelens[1]),
    as.character(outputDir))

  cat("done.\n")

  # produce summary.txt file

  summary <- read.delim(file.path(paths[1],"summary.txt"), row.names=1)
  summary[,1] <-factor(summary[,1], levels=c(levels(summary[,1]), length(sampleSets)))
  summary[grep("Number of conditions", rownames(summary)) ,1] <- length(sampleSets)
  celfiles <- c()
  for(path in paths) {
    cursum <- read.delim(file.path(path,"summary.txt"), row.names=1)
    celfiles <- append(celfiles, as.character(cursum[grep("CEL files", rownames(cursum)),1]))
  }
  summary[,1] <-factor(summary[,1], levels=c(levels(summary[,1]), paste(celfiles,collapse=", ")))
  summary[ grep("CEL files", rownames(summary)) ,1] <- paste(celfiles, collapse=", ")
  
  rowindices <- c()
  for(r in 1:length(rownames(summary))) {
    if(length(grep("Number of conditions", rownames(summary)[r]))>0 ||
       length(grep("Number of probes", rownames(summary)[r]))>0 ||
       length(grep("Number of probesets", rownames(summary)[r]))>0 ||
       length(grep("Number of categories", rownames(summary)[r]))>0 ||
       length(grep("Subsampling interval", rownames(summary)[r]))>0 ||
       length(grep("Number of sampling sweeps", rownames(summary)[r]))>0 ||
       length(grep("CEL files", rownames(summary)[r]))>0) {
          rowindices <- append(rowindices,r)
    }
  }
  condensedsummary <-  data.frame(summary[rowindices,] , row.names = rownames(summary)[rowindices])
  colnames(condensedsummary) <-  colnames(summary)
  write.table(condensedsummary, file.path(outputDir,"summary.txt"),  quote=FALSE, sep="\t")

  cat("Finished combining MMBGX runs.\n")
  cat("Now run readSingle.mmbgx(\"",outputDir,"\") to read into R for further analysis.\n", sep="")
}




readOutput.mmbgx <- function(...) {
  args <- list(...)
  paths <- c()
  for(i in 1:length(args)){
    for(j in 1:length(args[[i]])){
      paths <- append(paths,args[[i]][j])
    }
  }
  if(length(paths) > 1){
    cat("Combining the output from", paste(paths, collapse=", "),".\n\n")
    PSids <- NULL
    for(path in paths){
      pis <- scan(file.path(path, "PSids.txt"), what="character", quiet=TRUE)
      PSids <- cbind(PSids,pis)
    }
    for(i in nrow(PSids)){
      if(any(PSids[i,1] != PSids[i,])) stop("Sorry, incompatible output directories: probeset IDs do not match.\n")
    }
  }
  res <- list()
  for(path in paths){
    out <- readSingle.mmbgx(path)
    if(length(res) == 0) res <- out
    else {
      for(i in 1:ncol(out$muave)) res$mu[[length(res$mu)+1]] <- out$mu[[i]]
      res$muave <- cbind(res$muave, out$muave)
    }
  }
  cat("Reading transcript IDs...")
  res$PSids <- scan(file.path(paths[1], "PSids.txt"), what="character")
  if(file.exists(file.path(paths[1], "geneIDs.txt"))) {
    cat("Reading gene IDs...")
    res$geneIDs <- scan(file.path(paths[1], "geneIDs.txt"), what="character")
  }
  if(file.exists(file.path(path, "geneTranscripts.txt"))) {
    cat("Reading gene-transcript mappings...")
    res$geneTranscripts <- readVarLines(file.path(paths[1], "geneTranscripts.txt"))
  }
  cat("Done.\n")
  return(res)
}

loessNorm <- function (bgxOutput) {
  n.conds <- ncol(bgxOutput$muave)
  dum <- apply(bgxOutput$muave, 1, mean)
  for (k in 1:n.conds) {
    cat("Loess normalising condition ", k, "/", n.conds,"\r")
    lf <- loess(bgxOutput$muave[, k] - dum ~ dum)
    .Call("addByColumn", bgxOutput$muave, -lf$fitted,
      as.integer(nrow(bgxOutput$muave)), as.integer(k-1), as.integer(k-1))
    .Call("addByColumn", bgxOutput$mu[[k]], - lf$fitted,
      as.integer(nrow(bgxOutput$mu[[k]])), as.integer(0), as.integer(ncol(bgxOutput$mu[[k]]) -1))
  }
  cat("\nDone.\n")
}
 
### Plot densities of mu's under all conditions
plotExpressionDensity <- function(bgxOutput, probeset=NULL, ...) {
  if(is.null(probeset)) stop("Please specify a probeset index or ID.")
  if(is.character(probeset)) probeset<- (1:length(bgxOutput$PSids))[probeset == bgxOutput$PSids]

  noOfConditions <- length(bgxOutput$mu)
  ymax <- 0
  xmin <- 9999
  xmax <- -9999
  densities <- list(length=noOfConditions)
  
  for(i in 1:noOfConditions) {
    densities[[i]] <- density(bgxOutput$mu[[i]][probeset,], ...)
    yma <- max(densities[[i]]$y)
    xmi <- min(densities[[i]]$x)
    xma <- max(densities[[i]]$x)
    if(ymax < yma) ymax <- yma
    if(xmax < xma) xmax <- xma
    if(xmin > xmi) xmin <- xmi
  }

  plot(NA, type="n", xlab="Expression", ylab="Density", main=paste("Densities of mu for probeset", bgxOutput$PSids[probeset]), xlim=c(xmin,xmax), ylim=c(0,ymax+0.2))# leave some space for legend

  for(i in 1:noOfConditions) lines(densities[[i]], lty=i)

  legend("topright", paste("Cond", c(1:noOfConditions)), lty=c(1:noOfConditions), cex=0.75)

}

### Plot densities of differential expression between two conditions
plotDEDensity <- function(bgxOutput, probeset=NULL, conditions=c(1,2), ...) {
  if(is.null(probeset)) stop("Please specify a probeset index or ID.")
  if(is.character(probeset)) probeset <- (1:length(bgxOutput$PSids))[probeset == bgxOutput$PSids]
  if(length(conditions) !=2 ) stop("Please specify exactly two conditions (e.g. conditions=c(1,2)).")

  diff <- bgxOutput$mu[[conditions[2]]][probeset,] - bgxOutput$mu[[conditions[1]]][probeset,] 
  
  plot(density(diff, ...), main=paste("Density of mu for probeset", bgxOutput$PSids[probeset], "\ncondition ", conditions[2] , " - condition", conditions[1]), xlab="Differential Expression")
  abline(v=0)

}

### Rank probesets by differential expression
rankByDE <- function(bgxOutput, conditions=c(1,2), absolute=TRUE) {

  mu_diff <- bgxOutput$mu[[conditions[2]]] - bgxOutput$mu[[conditions[1]]]
  de <- c()
  tau <- var <- m <- 0
  for(i in 1:nrow(bgxOutput$mu[[1]])) {
    temp <- mu_diff[i,]

    # using sokal function
    var <- tau <- 0.0
    m <- 0
    sok <- .C("sokal", as.integer(1024), as.double(temp), as.double(var), as.double(tau), as.integer(m), PACKAGE="mmbgx")
    mcse <- sqrt(1023*sok[[3]]*sok[[4]]/(1024*(1024-sok[[4]])))

    # slow
    #tau <- 1
    #a <- acf(temp, plot=FALSE)
    #maxlag <- max(a$lag)
    #for(j in 2:maxlag) tau <- tryCatch(tau + 2 * (1-a$lag[j]/maxlag) * a$acf[j], error= function(e) NaN)
    #mcse <- sqrt((a$n.used-1)*tau*var(temp)/(a$n.used*(a$n.used-tau)))
    
    if(absolute) de[i] <- abs(mean(mu_diff[i,]))/mcse
    else de[i] <- mean(mu_diff[i,])/mcse
  }

  de[which(is.nan(de))] <- max(de, na.rm=T) # Get rid of NaNs returned in sokal()
  order <- sort(de,decreasing=TRUE, index.return=TRUE)$ix

  return(matrix(c(order,de[order]), dimnames=list( c(bgxOutput$PSids[order]), c("Position", "DiffExpression")), ncol=2))
}

### Plot sorted 2.5-97.5% CI of DE (not absolute value)
plotDiffRank <- function(bgxOutput, conditions=c(1,2), ymax=NULL) {

  mu_diff <- bgxOutput$mu[[conditions[2]]] - bgxOutput$mu[[conditions[1]]]
  
  if(is.null(ymax)) ymax <- nrow(mu_diff)
  
  quantiles <- matrix(ncol=2, nrow=nrow(mu_diff))
  for(r in 1:nrow(mu_diff)){
    q <- quantile(mu_diff[r,], probs=c(0.025,0.975))
    quantiles[r,1] <- q[[1]]
    quantiles[r,2] <- q[[2]]
   }
  
  plot(0,type="n", xlim=c(min(quantiles),max(quantiles)),ylim=c(0,ymax), xlab="CI 2.5%, 97.5%", ylab="Ranked probeset index")
  
  diff <- c()
  for(i in 1:nrow(bgxOutput$mu[[1]])) diff[i] <- mean(mu_diff[i,])/sd(mu_diff[i,])
  order <- sort(diff,decreasing=TRUE, index.return=TRUE)$ix
  
  for(r in 1:ymax) segments(quantiles[order[r],1],r, quantiles[order[r],2],r)

  abline(v=0, col=2)
  
}

### Estimate proportion of DE transcripts. Method similar to Efron
plotDEHistogram <- function(bgxOutput, conditions=c(1,2), df = floor(1.8 * log10(length(bgxOutput[["PSids"]])))) {

  diff <- bgxOutput$mu[[conditions[2]]] - bgxOutput$mu[[conditions[1]]]
  pp <- c()
  for(i in 1:length(bgxOutput$PSids)) pp[i] <- mean(diff[i,] < 0)

  if(length(pp) < 100) stop("Sorry, your sample of probesets is too small. This method only works well with 100 or more probesets.\n")

  cat("Degrees of freedom: ",df, "(decrease for smoother curve).\n")

#  require("splines")
  ppp <- pmax(pmin(pp, 1), 0)
  breaks <- seq(0, 1, by=1/ifelse(length(pp)<1000,20,40))
  ph <- hist(ppp, breaks = breaks, plot = FALSE)

  # fitting curve 'f' to histogram:

  x <- (breaks[-1] + breaks[ - length(breaks)])/2
  y <- ph$counts
  N <- length(y)
  f <- glm(y ~ poly(x, df = df), poisson)$fit

  # identifying local maxima and minima on curve 'f': 

  help=sign(f[1:(length(f)-1)]-f[2:(length(f))])
 
  i = 1
  signChanges=0
  numberSignChanges = 0
  localMaxima = 0
  numberLocalMaxima = 0
  localMinima = 0
  numberLocalMinima = 0

  if( help[1] > 0.0 ) {
    numberLocalMaxima = numberLocalMaxima+1
    localMaxima[numberLocalMaxima] = 1
  } else {
    numberLocalMinima = numberLocalMinima+1
    localMinima[numberLocalMinima] = 1
  }
  while(i < (length(f)-1)){
    if( sign(help[i]) != sign(help[i+1]) ){
       if( sign(help[i]) < 0.0 ){
         numberLocalMaxima = numberLocalMaxima+1
         localMaxima[numberLocalMaxima] = i+1
       } else{
         numberLocalMinima = numberLocalMinima+1
         localMinima[numberLocalMinima] = i+1
       }
       numberSignChanges = numberSignChanges+1
       signChanges[numberSignChanges] = i+1
    } 
    i = i+1
  }
  if( sign(help[(length(f)-1)]) < 0.0 ){
    numberLocalMaxima = numberLocalMaxima+1
    localMaxima[numberLocalMaxima] = length(f)
  } else{
    numberLocalMinima = numberLocalMinima+1
    localMinima[numberLocalMinima] = length(f)
  }

  specialPoints=union(union(1,signChanges),length(f))

  innermaxval=max(f[setdiff(localMaxima,c(1,40))])
  innermaxentry=localMaxima[f[localMaxima]==innermaxval]
 
  # fitting two curves to histogram: one to the left (f0left) and one to the right
  # (f0right) of the inner max. These curves make up the null.

  # fitting the curve left of innermax:
  #   - first fixing count y[1] to zero and giving it weight 1:

  y <- ph$counts    
  y[1]=0
  weights0=rep(1,times=length(y))

  #   - giving weight 0 to categories between the first and the minimum right of first:

  if( localMinima[1] > 1 ){
    weights0[2:(localMinima[1])]=0
  }

  # Abort if the innermaxentry is too near the edges
  if(length(1:innermaxentry) < df || length(innermaxentry:length(y)) < df) stop("Sorry, differential expression appears to be systematically biased. Maybe you need to normalise the data (see help(\"loessNorm\")) or maybe you are not using a representative subset of probesets.\n")

  f0left <- glm(y[1:innermaxentry] ~ poly(x[1:innermaxentry], df = df),weights=weights0[1:innermaxentry], poisson)$fit

  NumberNullGenesf0left=round(sum(f0left),0)

  # fitting the curve right of innermax:
  #   - first fixing count y[length(y)] to zero and giving it weight 1: 

  y <- ph$counts    
  y[length(y)]=0
  weights0=rep(1,times=length(y))

  #   - giving weight 0 to categories between last and minimum left of last category

  if( localMinima[length(localMinima)] < length(y) ){
    weights0[localMinima[length(localMinima)]:(length(y)-1)]=0 
  }

  f0right <- glm(y[innermaxentry:length(y)] ~ poly(x[innermaxentry:length(y)], df = df),weights=weights0[innermaxentry:length(y)], poisson)$fit
  
  NumberNullGenesf0right=round(sum(f0right[2:length(f0right)]),0)
 
  NumberNullGenesf0=min(NumberNullGenesf0left+NumberNullGenesf0right,sum(ph$counts))

  # taking care of roundings:

  NumberDiffGenesLeft=max(sum(ph$counts[1:innermaxentry])-NumberNullGenesf0left,0)
  NumberDiffGenesRight=max(sum(ph$counts[(innermaxentry+1):length(ph$counts)])-NumberNullGenesf0right,0)
  NumberDiffGenes=NumberDiffGenesLeft+NumberDiffGenesRight

  cat("Number of differentially expressed transcripts (left): ",NumberDiffGenesLeft,"\n")
  cat("Number of differentially expressed  transcripts (right): ",NumberDiffGenesRight,"\n")
  cat("Total number of differentially expressed transcripts: ",NumberDiffGenes,"\n")
  
  # plotting the histogram, fitted curve and intersting points:

  plot(ph,main=substitute(paste("Histogram of P(",mu[two] - mu[one]," < 0); df=",df),list(two=paste("g",conditions[2],sep=""),one=paste("g",conditions[1],sep=""),df=df)) ,xlab=paste("#DEG:",NumberDiffGenes," up:",NumberDiffGenesLeft," down:",NumberDiffGenesRight,sep=""))
  lines(x,f,col=8,lwd=2)
  points(x[localMaxima],f[localMaxima],col=2,pch=1,cex=2)  # maxima: red circle
  points(x[localMinima],f[localMinima],col=4,pch=2,cex=2)  # minima: blue
  points(x[innermaxentry],f[innermaxentry],col=1,pch=20,cex=2)
  lines(x[1:innermaxentry],f0left,col=1,lwd=2)
  lines(x[innermaxentry:length(y)],f0right,col=1,lwd=2)

#  return(c(NumberDiffGenesLeft,NumberDiffGenesRight))
}


plotTranscriptDensities <- function(resT, gene, filter=F, conditions=c(1,2), xlim=NULL, ylim=NULL) {
  g_ind <- ifelse(is.character(gene), which(resT$geneIDs==gene), gene)
  t_ids <- resT$geneTranscripts[[g_ind]]
  
  # check this is a good gene
  for(t in t_ids) {
    if(length(grep(gsub("\\+","\\|",resT$PSids[grep(t, resT$PSids)]), resT$geneTranscripts)) > 1) {
      stop("Aborting: gene contains cross-gene transcript-probesets.\n")
    }
  }
  
  t_ids4leg <- c()
  densities1 <- list()
  densities2 <- list()
  xmin <- 15
  xmax <- 0
  ymin <- 100
  ymax <- 0
  count <- 0
  for(t_ind in grep(paste(t_ids, collapse="|"), resT$PSids)) {
    fil <- F
    if(filter && isfiltered(resT, t_ind, conditions)) {
      fil <- T
      warning("Filtered ",resT$PSids[t_ind],".\n")
    }
    if(!fil) {
      count <- count+1
      t_ids4leg <- append(t_ids4leg, resT$PSids[t_ind])
      densities1[[count]] <- density(resT$mu[[conditions[1]]][t_ind,])
      densities2[[count]] <- density(resT$mu[[conditions[2]]][t_ind,])
      xmin <- ifelse(min(min(densities1[[count]]$x),min(densities2[[count]]$x)) < xmin,
                    min(min(densities1[[count]]$x),min(densities2[[count]]$x)), xmin) 
      xmax <- ifelse(max(max(densities1[[count]]$x),max(densities2[[count]]$x)) > xmax,
                    max(max(densities1[[count]]$x), max(densities2[[count]]$x)), xmax)
      ymin <- ifelse(min(min(densities1[[count]]$y),min(densities2[[count]]$y)) < ymin,
                    min(min(densities1[[count]]$y),min(densities2[[count]]$y)), ymin) 
      ymax <- ifelse(max(max(densities1[[count]]$y),max(densities2[[count]]$y)) > ymax,
                    max(max(densities1[[count]]$y),max(densities2[[count]]$y)), ymax) 
    }
  }
  if(count==0) stop("No transcripts to plot.")
  
  if(is.null(xlim)) xlim <- c(xmin, xmax)
  if(is.null(ylim)) ylim <- c(ymin, ymax+1)

  plot(0,type="n",xlim=xlim, ylim=ylim, ylab="Density",
          xlab="Log expression", main=paste("Transcript Densities (", resT$geneIDs[g_ind],")\nBlack: condition ",
                    conditions[1], ", red: condition ", conditions[2],sep=""))
  count <- 0
  for(i in 1:length(densities1)) {
    count <- count+1
    lines(densities1[[i]], col=1, lty=count)
    lines(densities2[[i]], col=2, lty=count)
  }
  legend("topright", t_ids4leg, lty=c(1:count))
}


differentialSplicing <- function(resG, resT, conditions=c(1,2)) {
  if(length(resG$PSids) > length(resT$PSids))
    stop("You have probably passed the gene and transcript-level objects in the wrong order.")

  ts <- t2ps(resT$PSids)
  
  t2g <- vector("list", sum(unlist(lapply(resT$geneTranscripts, length))))
  j <- 1
  for(i in 1:length(resT$geneTranscripts)) {
    t2g[j:(j+length(resT$geneTranscripts[[i]])-1)] <- i
    j <- j+length(resT$geneTranscripts[[i]])
  }
  names(t2g) <- unlist(resT$geneTranscripts)
  inditrans <- lapply(resT$PSids[grep("\\+", resT$PSids)], function(x) {strsplit(x,"\\+") })

  cat("Generating exclusion list of genes of with cross-gene transcript-probesets...")
  badgenes <- c() # genes with transcripts indistinguishable from transcripts in other genes
  for(i in inditrans) {
    ps <-  unlist(t2g[i[[1]]])
    if(any(ps!=ps[1])) {
      badgenes <- append(badgenes,unique(ps))
    }
  }
  badgenes <- unique(badgenes)
  cat("done.\n")
  
  result <- data.frame()

  cat("Searching for candidate genes for differential splicing:\n")
  for(g in 1:length(resT$geneTranscripts)) {
    if(length(resT$geneTranscripts[[g]]) <= 1) next()
    if(g %in% badgenes) next()
    
    g_ind <- which(resG$PSids == resT$geneIDs[g])
    g_diff <- resG$mu[[conditions[2]]][g_ind, ] - resG$mu[[conditions[1]]][g_ind, ]
    maxpg0 <- .5
    maxpg0diff <- 0
    allfiltered <- TRUE
    numberfiltered <- 0
    visited <- c() # keep track of visited probesets, do not iterate twice over
                   # same PS even if it corresponds to multiple transcripts

    maxt_index <- c() # t with greatest pg0

    for(t in resT$geneTranscripts[[g]]) {
      #t_ind <- grep(t, resT$PSids)
      t_ind <- ts[[t]]
      if(length(t_ind)==0) {
        cat("Warning: ", t , " has no probes.\n")
        next()
      }
 
      if(t_ind %in% visited) next()
      visited <- append(visited, t_ind)

      t_diff <- resT$mu[[conditions[2]]][t_ind, ] - resT$mu[[conditions[1]]][t_ind, ]
      #pd <- sum(t_diff > 0)/1024
      #pd <- max(pd,1-pd)
      filter <- FALSE
      if(isfiltered(resT, t_ind, conditions)) {
        filter <- TRUE
        numberfiltered <- numberfiltered+1
      }
      
      if(!filter) allfiltered <- FALSE
      if(!filter && abs((sum(t_diff - g_diff > 0 )/1024) - .5) > abs(maxpg0-.5)) {
        maxpg0 <- sum(t_diff - g_diff > 0 )/1024
        maxpg0diff <- mean(t_diff-g_diff) 
        maxt_index <- t_ind
      }
    }
    if(!allfiltered && length(visited)-numberfiltered > 1) {
      result <- rbind(result, data.frame(cbind(
                        g_index=g,
                        g_maxpg0=maxpg0,
                        g_maxdiff=maxpg0diff,
                        t_index=maxt_index,
                        g_transconsidered=length(visited)-numberfiltered)))
    }
    #cat("Warning: all filtered for ", resG$PSids[g], '\n')
    cat(g,'\r')
  }
  return(result)
}


readVarLines <- function(file, sep="\t") {
  text <- readLines(file)
  res <- list()
  for(i in 1:length(text)) {
    r <- strsplit(text[i], sep)[[1]]
    res[[i]] <- r;
   
  }
  return(res)
}

isfiltered <- function(resT, transcript, conditions=c(1,2), pdthres=0.95) {
  t_ind <- ifelse(is.character(transcript), which(resT$PSids==transcript), transcript)
  t_diff <- resT$mu[[conditions[2]]][t_ind, ] - resT$mu[[conditions[1]]][t_ind, ]
  pd <- sum(t_diff < 0)/1024
  pd <- max(pd,1-pd)
  if(resT$muave[t_ind,conditions[1]] < 1.5 && resT$muave[t_ind,conditions[2]] < 1.5 && pd < pdthres)
     return(TRUE)
  else return(FALSE)
}


# create hash table for each transcript linking to index in PSids 
t2ps <- function(PSids) {
  ts <- list()
  wh <- grep("+", PSids, fixed=T)  
    
  tcount <- 1
  pscount <- 1
  for(i in wh) {
    ts[tcount:(tcount+i-pscount-1)] <- c(pscount:(i-1))
    tcount <- tcount+i-pscount
    for(s in strsplit(PSids[i], "+", fixed=TRUE)[[1]]) {
      ts[[tcount]] <- i
      tcount <- tcount+1
    } 
    pscount <- i+1 
  }

  tn <- unlist(strsplit(PSids, "+", fixed=TRUE))
  ts[tcount:length(tn)] <- c(pscount:length(PSids))
  names(ts) <- unlist(strsplit(PSids, "+", fixed=TRUE))
  return(ts)
}


